import pytest
import os
from fluent_tfx import PipelineDef


@pytest.fixture
def pipeline_def(tmp_path):
    bucket_uri = os.path.join(tmp_path.absolute(), 'bucket')
    return PipelineDef('pipeline_def_name', bucket=bucket_uri)



import pytest

import os
from typing import Text

import tensorflow as tf

from examples.chicago_taxi_pipeline.pipeline import get_pipeline
from fluent_tfx.pipeline_def import PipelineDef
from tfx.orchestration import metadata
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner


@pytest.mark.skip('default example code from tfx broken?')
class ChicagoPipelineFTFXEndToEndTest(tf.test.TestCase):

    def setUp(self):
        super(ChicagoPipelineFTFXEndToEndTest, self).setUp()

        self._bucket_dir = os.path.join(os.path.dirname(__file__), 'bucket')
        self._initial_pipeline_def = PipelineDef(
            name='chicago_pipeline_e2e', bucket=self._bucket_dir)

    def assertComponentExecutedOnce(self, component: Text) -> None:
        component_path = os.path.join(
            self._bucket_dir, self._initial_pipeline_def.pipeline_name, component)
        self.assertTrue(tf.io.gfile.exists(component_path))
        outputs = tf.io.gfile.listdir(component_path)
        for output in outputs:
            execution = tf.io.gfile.listdir(
                os.path.join(component_path, output))
            self.assertEqual(1, len(execution))

    def assertPipelineExecution(self) -> None:
        self.assertExecutedOnce('CsvExampleGen')
        self.assertExecutedOnce('StatisticsGen')
        self.assertExecutedOnce('SchemaGen')
        self.assertExecutedOnce('Transform')
        self.assertExecutedOnce('Tuner')
        self.assertExecutedOnce('Trainer')
        self.assertExecutedOnce('Evaluator')
        self.assertExecutedOnce('Pusher')

    def testChicagoE2EFTFXPipeline(self):
        pipeline_def = get_pipeline(self._initial_pipeline_def)
        pipeline_def = pipeline_def.with_sqlite_ml_metadata().cache()

        pipeline = pipeline_def.build()
        BeamDagRunner().run(pipeline)

        artifact_root = os.path.join(
            self._bucket_dir, pipeline_def.pipeline_name)
        self.assertTrue(tf.io.gfile.exists(
            os.path.join(artifact_root, 'serving_model')))
        self.assertTrue(os.path.join(artifact_root, 'metadata.db'))
        expected_execution_count = 8  # 7 components + 1 resolver
        metadata_config = pipeline_def.metadata_connection_config
        with metadata.Metadata(metadata_config) as m:
            artifact_count = len(m.store.get_artifacts())
            execution_count = len(m.store.get_executions())
            self.assertGreaterEqual(artifact_count, execution_count)
            self.assertEqual(expected_execution_count, execution_count)

        self.assertPipelineExecution()

        BeamDagRunner().run(pipeline)

        # All executions but Evaluator and Pusher are cached.
        with metadata.Metadata(metadata_config) as m:
            # Artifact count is increased by 3 caused by Evaluator and Pusher.
            self.assertEqual(artifact_count + 3, len(m.store.get_artifacts()))
            artifact_count = len(m.store.get_artifacts())
            self.assertEqual(expected_execution_count * 2,
                             len(m.store.get_executions()))

        BeamDagRunner().run(pipeline)

        # Asserts cache execution.
        with metadata.Metadata(metadata_config) as m:
            # Artifact count is unchanged.
            self.assertEqual(artifact_count, len(m.store.get_artifacts()))
            self.assertEqual(expected_execution_count * 3,
                             len(m.store.get_executions()))


if __name__ == '__main__':
    tf.compat.v1.enable_v2_behavior()
    tf.test.main()



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_pipeline_def_from_csv_works(pipeline_def, monkeypatch):
    # Arrange
    mock_from_csv = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.from_csv', mock_from_csv)

    # Act
    pipeline_def = pipeline_def.from_csv(
        uri='some_uri', input_config=1, output_config=2)
    _, kwargs = mock_from_csv.call_args

    # Assert
    assert mock_from_csv.called
    assert kwargs['uri'] == 'some_uri'
    assert kwargs['input_config'] == 1
    assert kwargs['output_config'] == 2

    assert pipeline_def.components['example_gen'] is mock_from_csv.return_value
    assert pipeline_def.example_gen is mock_from_csv.return_value


def test_pipeline_def_from_tfrecord_works(pipeline_def, monkeypatch):
    # Arrange
    mock_from_tfrecord = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.from_tfrecord', mock_from_tfrecord)

    # Act
    pipeline_def = pipeline_def.from_tfrecord(
        uri='some_uri', input_config=1, output_config=2)
    _, kwargs = mock_from_tfrecord.call_args

    # Assert
    assert mock_from_tfrecord.called
    assert kwargs['uri'] == 'some_uri'
    assert kwargs['input_config'] == 1
    assert kwargs['output_config'] == 2

    assert pipeline_def.components['example_gen'] is mock_from_tfrecord.return_value
    assert pipeline_def.example_gen is mock_from_tfrecord.return_value


def test_pipeline_def_from_bigquery_works(pipeline_def, monkeypatch):
    # Arrange
    mock_from_bigquery = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.from_bigquery', mock_from_bigquery)

    # Act
    pipeline_def = pipeline_def.from_bigquery(
        query='some_query', input_config=1, output_config=2)
    _, kwargs = mock_from_bigquery.call_args

    # Assert
    assert mock_from_bigquery.called
    assert kwargs['query'] == 'some_query'
    assert kwargs['input_config'] == 1
    assert kwargs['output_config'] == 2

    assert pipeline_def.components['example_gen'] is mock_from_bigquery.return_value
    assert pipeline_def.example_gen is mock_from_bigquery.return_value


def test_pipeline_def_from_custom_example_gen_component_works(pipeline_def):
    # Arrange
    mock_base_component = mock.MagicMock(spec=BaseComponent)

    # Act
    pipeline_def = pipeline_def.from_custom_example_gen_component(
        mock_base_component)

    # Assert
    assert pipeline_def.components['example_gen'] is mock_base_component
    assert pipeline_def.example_gen is mock_base_component



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_pipeline_def_with_base_model_works(pipeline_def, monkeypatch):
    # Arrange
    mock_with_base_model = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.with_base_model', mock_with_base_model)

    # Act
    pipeline_def = pipeline_def.with_base_model('some_uri')
    _, kwargs = mock_with_base_model.call_args

    # Assert
    assert mock_with_base_model.called
    assert kwargs['uri'] == 'some_uri'
    assert pipeline_def.components['train_base_model'] is mock_with_base_model.return_value
    assert pipeline_def.train_base_model is mock_with_base_model.return_value


def test_pipeline_def_train_minimally_works(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_transform = mock.MagicMock()
    pipeline_def.transform = mock_transform

    pipeline_def.cached_example_input = None

    pipeline_def.tuner = None
    pipeline_def.user_hyperparameters_importer = None

    mock_trainer = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Trainer', mock_trainer)

    mock_module = mock.Mock()

    # Act
    pipeline_def = pipeline_def.train(module_file=mock_module)
    _, kwargs = mock_trainer.call_args

    # Assert
    assert mock_trainer.called
    assert 'train_args' not in kwargs
    assert 'eval_args' not in kwargs
    assert kwargs['schema'] is SchemaInputs.SCHEMA_CHANNEL(pipeline_def)
    assert kwargs['transform_graph'] is mock_transform.outputs['transform_graph']
    assert 'custom_config' not in kwargs
    assert 'hyperparameters' not in kwargs
    assert isinstance(kwargs['custom_executor_spec'],
                      executor_spec.ExecutorClassSpec)
    assert kwargs['transformed_examples'] is ExampleInputs.PREPROCESSED_EXAMPLES(
        pipeline_def)
    assert pipeline_def.components['trainer'] is mock_trainer.return_value
    assert pipeline_def.trainer is mock_trainer.return_value


def test_pipeline_def_train_works_with_optional_args(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_hparams = mock.MagicMock()
    pipeline_def.user_hyperparameters_importer = mock_hparams

    mock_transform = mock.MagicMock()
    pipeline_def.transform = mock_transform

    mock_example_input = mock.MagicMock()
    pipeline_def.example_gen = mock_example_input
    pipeline_def.cached_example_input = None

    pipeline_def.tuner = None

    mock_trainer = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Trainer', mock_trainer)

    mock_module = mock.Mock()
    mock_config = mock.Mock()
    mock_executor_spec = mock.Mock()

    # Act
    pipeline_def = pipeline_def.train(
        module_file=mock_module,
        train_args=1, eval_args=2,
        custom_config=mock_config,
        custom_executor_spec=mock_executor_spec,
        example_input=ExampleInputs.RAW_EXAMPLES
    )
    _, kwargs = mock_trainer.call_args

    # Assert
    assert mock_trainer.called
    assert kwargs['train_args'] == 1
    assert kwargs['eval_args'] == 2
    assert kwargs['transformed_examples'] is ExampleInputs.RAW_EXAMPLES(
        pipeline_def)
    assert pipeline_def.cached_example_input is ExampleInputs.RAW_EXAMPLES(
        pipeline_def)
    assert kwargs['schema'] is SchemaInputs.SCHEMA_CHANNEL(pipeline_def)
    assert kwargs['transform_graph'] is mock_transform.outputs['transform_graph']
    assert kwargs['custom_config'] is mock_config
    assert kwargs['custom_executor_spec'] is mock_executor_spec
    assert kwargs['hyperparameters'] is HyperParameterInputs.BEST_HYPERPARAMETERS(
        pipeline_def)
    assert pipeline_def.components['trainer'] is mock_trainer.return_value
    assert pipeline_def.trainer is mock_trainer.return_value


def test_pipeline_def_train_uses_cached_example_input(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    pipeline_def.transform = None

    mock_input = mock.MagicMock()
    pipeline_def.cached_example_input = mock_input

    pipeline_def.tuner = None
    pipeline_def.user_hyperparameters_importer = None

    mock_trainer = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Trainer', mock_trainer)

    mock_module = mock.Mock()

    # Act
    pipeline_def = pipeline_def.train(module_file=mock_module)
    _, kwargs = mock_trainer.call_args

    # Assert
    assert mock_trainer.called
    assert kwargs['transformed_examples'] is mock_input
    assert pipeline_def.components['trainer'] is mock_trainer.return_value
    assert pipeline_def.trainer is mock_trainer.return_value


def test_pipeline_def_train_can_use_example_inputs(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    pipeline_def.transform = None

    mock_example_gen = mock.MagicMock()
    pipeline_def.example_gen = mock_example_gen
    pipeline_def.cached_example_input = None

    pipeline_def.tuner = None
    pipeline_def.user_hyperparameters_importer = None

    mock_trainer = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Trainer', mock_trainer)

    mock_module = mock.Mock()

    # Act
    pipeline_def = pipeline_def.train(
        module_file=mock_module, example_input=ExampleInputs.RAW_EXAMPLES)
    _, kwargs = mock_trainer.call_args

    # Assert
    assert mock_trainer.called
    assert kwargs['transformed_examples'] is ExampleInputs.RAW_EXAMPLES(
        pipeline_def)
    assert pipeline_def.components['trainer'] is mock_trainer.return_value
    assert pipeline_def.trainer is mock_trainer.return_value


def test_pipeline_def_train_updates_config_with_ai_platform_args(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_transform = mock.MagicMock()
    pipeline_def.transform = mock_transform

    pipeline_def.cached_example_input = None

    pipeline_def.tuner = None
    pipeline_def.user_hyperparameters_importer = None

    mock_trainer = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Trainer', mock_trainer)

    mock_module = mock.Mock()

    ai_platform_args = mock.Mock()
    custom_config = mock.Mock()
    # Act
    pipeline_def = pipeline_def.train(
        module_file=mock_module, ai_platform_args=ai_platform_args, custom_config=custom_config)
    _, kwargs = mock_trainer.call_args

    # Assert
    assert mock_trainer.called
    assert kwargs['custom_config'][ai_platform_trainer_executor.TRAINING_ARGS_KEY] is ai_platform_args
    assert pipeline_def.components['trainer'] is mock_trainer.return_value
    assert pipeline_def.trainer is mock_trainer.return_value



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_infra_validator_works(pipeline_def, monkeypatch):
    # Arrange
    mock_infra_validator = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.InfraValidator', mock_infra_validator)

    mock_example_gen = mock.MagicMock()
    pipeline_def.example_gen = mock_example_gen

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_serving_spec = mock.Mock()

    # Act
    pipeline_def = pipeline_def.infra_validate(serving_spec=mock_serving_spec)
    _, kwargs = mock_infra_validator.call_args

    # Assert
    assert mock_infra_validator.called
    assert kwargs['serving_spec'] is mock_serving_spec
    assert 'validation_spec' not in kwargs
    assert 'request_spec' not in kwargs
    assert kwargs['model'] is mock_trainer.outputs['model']
    assert kwargs['examples'] is ExampleInputs.RAW_EXAMPLES(pipeline_def)
    assert pipeline_def.components['infra_validator'] is mock_infra_validator.return_value
    assert pipeline_def.infra_validator is mock_infra_validator.return_value


def test_infra_validator_passes_works_with_optional_args(pipeline_def, monkeypatch):
    # Arrange
    mock_infra_validator = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.InfraValidator', mock_infra_validator)

    mock_transform = mock.MagicMock()
    pipeline_def.transform = mock_transform

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_serving_spec = mock.Mock()
    mock_validation_spec = mock.Mock()
    mock_request_spec = mock.Mock()

    # Act
    pipeline_def = pipeline_def.infra_validate(
        serving_spec=mock_serving_spec,
        validation_spec=mock_validation_spec,
        request_spec=mock_request_spec,
        example_input=ExampleInputs.PREPROCESSED_EXAMPLES)
    _, kwargs = mock_infra_validator.call_args

    # Assert
    assert mock_infra_validator.called
    assert kwargs['request_spec'] is mock_request_spec
    assert kwargs['validation_spec'] is mock_validation_spec
    assert kwargs['examples'] is ExampleInputs.PREPROCESSED_EXAMPLES(
        pipeline_def)
    assert pipeline_def.components['infra_validator'] is mock_infra_validator.return_value
    assert pipeline_def.infra_validator is mock_infra_validator.return_value


def test_infra_validator_can_use_cached_input(pipeline_def, monkeypatch):
    # Arrange
    mock_infra_validator = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.InfraValidator', mock_infra_validator)

    mock_cached_input = mock.Mock()
    pipeline_def.cached_example_input = mock_cached_input

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_serving_spec = mock.Mock()

    # Act
    pipeline_def = pipeline_def.infra_validate(
        serving_spec=mock_serving_spec)
    _, kwargs = mock_infra_validator.call_args

    # Assert
    assert mock_infra_validator.called
    assert kwargs['examples'] is mock_cached_input
    assert pipeline_def.components['infra_validator'] is mock_infra_validator.return_value
    assert pipeline_def.infra_validator is mock_infra_validator.return_value



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor

from tfx.proto import pusher_pb2


def test_push_to_works(pipeline_def, monkeypatch):
    # Arrange
    mock_pusher = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Pusher', mock_pusher)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_push_dest = mock.Mock()

    pipeline_def.model_evaluator = None
    # Act
    pipeline_def = pipeline_def.push_to(push_destination=mock_push_dest)
    _, kwargs = mock_pusher.call_args

    # Assert
    assert mock_pusher.called
    assert kwargs['model'] is mock_trainer.outputs['model']
    assert kwargs['push_destination'] is mock_push_dest
    assert 'model_blessing' not in kwargs
    assert 'custom_config' not in kwargs
    assert 'custom_executor_spec' not in kwargs
    assert 'infra_blessing' not in kwargs
    assert pipeline_def.components['pusher'] is mock_pusher.return_value
    assert pipeline_def.pusher is mock_pusher.return_value


def test_push_to_works_with_optional_args(pipeline_def, monkeypatch):
    # Arrange
    mock_pusher = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Pusher', mock_pusher)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_evaluator = mock.MagicMock()
    pipeline_def.model_evaluator = mock_evaluator

    mock_infra_validator = mock.MagicMock()
    pipeline_def.infra_validator = mock_infra_validator

    mock_config = mock.Mock()
    mock_exec = mock.Mock()

    # Act
    pipeline_def = pipeline_def.push_to(
        relative_push_uri='some_uri',
        custom_config=mock_config,
        custom_executor_spec=mock_exec)
    _, kwargs = mock_pusher.call_args

    # Assert
    assert mock_pusher.called
    assert kwargs['model'] is mock_trainer.outputs['model']
    assert kwargs['model_blessing'] is mock_evaluator.outputs['blessing']
    assert kwargs['custom_config'] is mock_config
    assert kwargs['custom_executor_spec'] is mock_exec
    assert kwargs['infra_blessing'] is mock_infra_validator.outputs['blessing']

    assert isinstance(kwargs['push_destination'], pusher_pb2.PushDestination)
    assert(
        kwargs['push_destination'].filesystem.base_directory.endswith('some_uri'))

    assert pipeline_def.components['pusher'] is mock_pusher.return_value
    assert pipeline_def.pusher is mock_pusher.return_value



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor

from tfx.proto import pusher_pb2


def test_bulk_infer_works(pipeline_def, monkeypatch):
    # Arrange
    mock_bulk_inferrer = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.BulkInferrer', mock_bulk_inferrer)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    pipeline_def.model_evaluator = None

    mock_example_provider_component = mock.MagicMock(spec=BaseComponent)

    # Act
    pipeline_def = pipeline_def.bulk_infer(
        example_provider_component=mock_example_provider_component)
    _, kwargs = mock_bulk_inferrer.call_args

    # Assert
    assert mock_bulk_inferrer.called
    assert kwargs['model'] is mock_trainer.outputs['model']
    assert 'model_blessing' not in kwargs
    assert 'data_spec' not in kwargs
    assert 'model_spec' not in kwargs

    assert pipeline_def.components['bulk_inferrer_example_provider'] is mock_example_provider_component

    assert pipeline_def.components['bulk_inferrer'] is mock_bulk_inferrer.return_value
    assert pipeline_def.bulk_inferrer is mock_bulk_inferrer.return_value


def test_bulk_infer_works_with_optional_args(pipeline_def, monkeypatch):
    # Arrange
    mock_bulk_inferrer = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.BulkInferrer', mock_bulk_inferrer)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_evaluator = mock.MagicMock()
    pipeline_def.model_evaluator = mock_evaluator

    mock_example_provider_component = mock.MagicMock(spec=BaseComponent)

    mock_data_spec = mock.Mock()
    mock_model_spec = mock.Mock()

    # Act
    pipeline_def = pipeline_def.bulk_infer(
        example_provider_component=mock_example_provider_component,
        data_spec=mock_data_spec,
        model_spec=mock_model_spec)
    _, kwargs = mock_bulk_inferrer.call_args

    # Assert
    assert mock_bulk_inferrer.called
    assert kwargs['model'] is mock_trainer.outputs['model']
    assert kwargs['model_blessing'] is mock_evaluator.outputs['blessing']
    assert kwargs['data_spec'] is mock_data_spec
    assert kwargs['model_spec'] is mock_model_spec

    assert pipeline_def.components['bulk_inferrer_example_provider'] is mock_example_provider_component

    assert pipeline_def.components['bulk_inferrer'] is mock_bulk_inferrer.return_value
    assert pipeline_def.bulk_inferrer is mock_bulk_inferrer.return_value



# MIT License

# Copyright(c) 2020 Theodoros Ntakouris

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import logging
from typing import Dict, Text, Any, Optional, List

import absl
import fluent_tfx as ftfx
import tensorflow as tf
import kerastuner
from kerastuner.tuners import Hyperband
import tensorflow_transform as tft
import tensorflow_model_analysis as tfma
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner

from tfx.dsl.component.experimental.decorators import component
from tfx.dsl.component.experimental.annotations import InputArtifact
from tfx.types.standard_artifacts import Model, HyperParameters, Schema

from tfx.components.trainer.fn_args_utils import FnArgs
from tfx.components.trainer.executor import TrainerFnArgs
from tfx.components.tuner.component import TunerFnResult

from tfx.proto import trainer_pb2, evaluator_pb2, pusher_pb2, infra_validator_pb2

from google.protobuf import text_format
from tensorflow.python.lib.io import file_io
from tensorflow_metadata.proto.v0 import schema_pb2
from tensorflow_transform.tf_metadata import schema_utils

from examples.usage_guide.model_code import LABEL_KEY

from tfx.dsl.component.experimental.decorators import component
from tfx.dsl.component.experimental.annotations import InputArtifact
from tfx.types.standard_artifacts import Model, HyperParameters, Schema

"""
This file contains code the pipeline orchestration that uses user provided code from
`model_code.py`. Only the data sources and evaluation configuration is declared with high fidelity
here, basically.
"""


@component
def tips_printer(model: InputArtifact[Model],
                 hyperparameters: InputArtifact[HyperParameters]) -> None:
    # just showcase that you can hook onto any pre-existing component
    logging.info(
        f'for next runs, you can use with_base_model({model.uri})')
    logging.info(
        f'for next runs, you can use with_hyperparameters({hyperparameters.uri})')


def tips_printer_build_fn(components):
    return tips_printer(hyperparameters=components['tuner'].outputs['best_hyperparameters'],
                        model=components['trainer'].outputs['model'])


def _get_eval_config() -> tfma.EvalConfig:
    return tfma.EvalConfig(
        model_specs=[tfma.ModelSpec(
            label_key=LABEL_KEY)],
        slicing_specs=[tfma.SlicingSpec()],
        metrics_specs=[
            tfma.MetricsSpec(metrics=[
                tfma.MetricConfig(
                    class_name='BinaryAccuracy',
                    threshold=tfma.MetricThreshold(
                        value_threshold=tfma.GenericValueThreshold(
                            lower_bound={'value': 0.01}),
                        change_threshold=tfma.GenericChangeThreshold(
                            direction=tfma.MetricDirection.HIGHER_IS_BETTER,
                            absolute={'value': 1e-10})))
            ])
        ])


def get_pipeline(pipeline_def: ftfx.PipelineDef) -> ftfx.PipelineDef:
    current_dir = os.path.dirname(
        os.path.realpath(__file__))

    user_code_file = os.path.join(os.path.dirname(
        os.path.realpath(__file__)), 'model_code.py')
    logging.info(
        f'Using {user_code_file} for preprocessing, training and tuning functions')

    return pipeline_def.from_csv(os.path.join(current_dir, 'data/')) \
        .generate_statistics() \
        .infer_schema(infer_feature_shape=True) \
        .validate_input_data() \
        .preprocess(user_code_file) \
        .tune(user_code_file,
              train_args=trainer_pb2.TrainArgs(num_steps=5),
              eval_args=trainer_pb2.EvalArgs(num_steps=3)) \
        .train(user_code_file,
               train_args=trainer_pb2.TrainArgs(num_steps=10),
               eval_args=trainer_pb2.EvalArgs(num_steps=5)) \
        .evaluate_model(eval_config=_get_eval_config()) \
        .infra_validate(serving_spec=infra_validator_pb2.ServingSpec(
            tensorflow_serving=infra_validator_pb2.TensorFlowServing(
                tags=['latest']),
            local_docker=infra_validator_pb2.LocalDockerConfig()
        ),
        request_spec=infra_validator_pb2.RequestSpec(
            tensorflow_serving=infra_validator_pb2.TensorFlowServingRequestSpec()
        )) \
        .push_to(relative_push_uri='serving') \
        .bulk_infer(example_provider_component=ftfx.input_builders.from_csv(
            uri=os.path.join(current_dir, 'to_infer'),
            name='bulk_infer_example_gen'
        )) \
        .add_custom_component(name='tips_printer', component=tips_printer_build_fn)


if __name__ == '__main__':
    absl.logging.set_verbosity(absl.logging.ERROR)

    bucket_uri = os.path.join(os.path.dirname(__file__), 'bucket')

    pipeline_def = ftfx.PipelineDef(name='simple_e2e', bucket=bucket_uri) \
        .with_sqlite_ml_metadata()

    pipeline_def = get_pipeline(pipeline_def)

    # you can also do:
    print('Exposed pipeline components dict:')
    print(pipeline_def.components)

    pipeline = pipeline_def.build()
    BeamDagRunner().run(pipeline)



from typing import Text, Any, Optional, List, Dict
from tfx.components import CsvExampleGen, BigQueryExampleGen, ImportExampleGen, ImporterNode, ResolverNode
from tfx.utils.dsl_utils import csv_input, tfrecord_input, external_input
from tfx.types import Channel, standard_artifacts
from tfx.dsl.experimental import latest_blessed_model_resolver
from tfx.proto import example_gen_pb2


def from_csv(uri: Text, name: Optional[Text] = None,
             input_config: Optional[example_gen_pb2.Input] = None,
             output_config: Optional[example_gen_pb2.Output] = None):
    """Constructs a CsvExampleGen component by using external_input(uri)

    Args:
        uri (Text): Csv file(s) uri
        name(Optional[Text]): The optional instance_name of the component. Please use this to avoid
        same-name errors.
        input_config (Optional[example_gen_pb2.Input], optional): Defaults to None.
        output_config (Optional[example_gen_pb2.Output], optional): Defaults to None.

    Returns: CsvExampleGen
    """
    args = {
        'input': external_input(uri),
    }

    if name:
        args['instance_name'] = name

    if input_config:
        args['input_config'] = input_config

    if output_config:
        args['output_config'] = output_config

    return CsvExampleGen(**args)


def from_tfrecord(uri: Text, name: Optional[Text] = None,
                  input_config: Optional[example_gen_pb2.Input] = None,
                  output_config: Optional[example_gen_pb2.Output] = None):
    """Constructs an ImportExampleGen component by using external_input(uri)

    Args:
        uri (Text): TFRecord file(s) uri
        input_config (Optional[example_gen_pb2.Input], optional): Defaults to None.
        output_config (Optional[example_gen_pb2.Output], optional): Defaults to None.

    Returns: ImportExampleGen
    """
    args = {
        'input': external_input(uri),
    }

    if name:
        args['instance_name'] = name

    if input_config:
        args['input_config'] = input_config

    if output_config:
        args['output_config'] = output_config

    return ImportExampleGen(**args)


def from_bigquery(query: Text, name: Optional[Text] = None,
                  input_config: Optional[example_gen_pb2.Input] = None,
                  output_config: Optional[example_gen_pb2.Output] = None):
    """Constructs a BigQueryExampleGen component by using external_input(uri)

    Args:
        query (Text): The query to run
        input_config (Optional[example_gen_pb2.Input], optional): Defaults to None.
        output_config (Optional[example_gen_pb2.Output], optional): Defaults to None.

    Returns: BigQueryExampleGen
    """
    args = {
        'query': query,
    }

    if name:
        args['instance_name'] = name

    if input_config:
        args['input_config'] = input_config

    if output_config:
        args['output_config'] = output_config

    return BigQueryExampleGen(**args)


def with_imported_schema(uri: Text):
    """Constructs an ImporterNode component that imports
    the schema in the pipelineas an artifact.

    If infer_schema is called, the subsequent components will still use this
    use provided schema, but the SchemaGen component will still produce inferred
    schema artifacts.

    Args:
        uri (Text): Schema .pbtxt file uri

    Returns: ImporterNode
    """
    return ImporterNode(
        instance_name='with_imported_schema',
        source_uri=uri,
        artifact_type=standard_artifacts.Schema)


def with_base_model(uri: Text):
    """Constructs an ImporterNode component that imports a `standard_artifacts.Model`
    artifact to use as a starting point for training.

    Args:
        uri (Text): Model artifact's uri

    Returns: ImporterNode
    """
    return ImporterNode(
        instance_name='with_base_model',
        source_uri=uri,
        artifact_type=standard_artifacts.Model)


def with_hyperparameters(uri: Text):
    """Constructs an ImporterNode component that imports a `standard_artifacts.HyperParameters`
    artifact to use for future runs.

    Args:
        uri (Text): Hyperparameter artifact's uri

    Returns: ImporterNode
    """
    return ImporterNode(
        instance_name='with_hyperparameters',
        source_uri=uri,
        artifact_type=standard_artifacts.HyperParameters)


def get_latest_blessed_model_resolver():
    """Constructs a latest blessed model resolver node

    Returns: ResolverNode
    """
    return ResolverNode(
        instance_name='latest_blessed_model_resolver',
        resolver_class=latest_blessed_model_resolver.LatestBlessedModelResolver,
        model=Channel(type=standard_artifacts.Model),
        model_blessing=Channel(type=standard_artifacts.ModelBlessing))



# MIT License

# Copyright(c) 2020 Theodoros Ntakouris

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
from operator import attrgetter

import absl
import fluent_tfx as ftfx
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner

from tfx.dsl.component.experimental.decorators import component
from tfx.dsl.component.experimental.annotations import InputArtifact
from tfx.types.standard_artifacts import Schema


@component
def schema_printer(schema_generated: InputArtifact[Schema],
                   schema_provided: InputArtifact[Schema]) -> None:
    print(f'uri of schema_generated: {schema_generated.uri}')
    print(f'uri of schema_provided: {schema_provided.uri}')


def schema_printer_build_fn(components):
    return schema_printer(schema_generated=components['schema_gen'].outputs['schema'],
                          schema_provided=components['user_schema_importer'].outputs['result'])


def get_pipeline(pipeline_def: ftfx.PipelineDef) -> ftfx.PipelineDef:
    current_dir = os.path.dirname(
        os.path.realpath(__file__))
    return pipeline_def \
        .from_csv(os.path.join(current_dir, 'data/')) \
        .generate_statistics() \
        .infer_schema() \
        .with_imported_schema(os.path.join(current_dir, 'saved', 'schema')) \
        .add_custom_component(name='schema_printer', component=schema_printer_build_fn)


if __name__ == '__main__':
    absl.logging.set_verbosity(absl.logging.ERROR)

    bucket_uri = os.path.join(os.path.dirname(__file__), 'bucket')
    pipeline_def = ftfx.PipelineDef(
        name='custom_component_schema', bucket=bucket_uri).with_sqlite_ml_metadata()

    pipeline_def = get_pipeline(pipeline_def)
    pipeline = pipeline_def.build()

    BeamDagRunner().run(pipeline)
